#!/usr/bin/env python2
#
# This file is provided under a dual BSD/GPLv2 license.  When using or
# redistributing this file, you may do so under either license.
#
# GPL LICENSE SUMMARY
#
# Copyright(c) 2012-2017 Intel Corporation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# Contact Information:
# email: cas_community@intel.com
# www:   www.intel.com/cas
#
# BSD LICENSE
#
# Copyright(c) 2012-2017 Intel Corporation.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of Intel Corporation nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE

import subprocess
import csv
import re

# Casadm functionality

class casadm:
    casadm_path = '/sbin/casadm'

    class result:
        def __init__(self, cmd):
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            self.exit_code = p.wait()
            output = p.communicate()
            self.stdout = output[0]
            self.stderr = output[1]

    class CasadmError(Exception):
        def __init__(self, result):
            super(casadm.CasadmError, self).__init__("casadm error")
            self.result = result

    @classmethod
    def run_cmd(cls, cmd):
        result = cls.result(cmd)
        if result.exit_code != 0:
            raise cls.CasadmError(result)
        return result

    @classmethod
    def list_caches(cls):
        cmd = [cls.casadm_path,
                    '--list-caches',
                    '--output-format', 'csv']
        return cls.run_cmd(cmd)

    @classmethod
    def check_cache_device(cls, device):
        cmd = [cls.casadm_path,
                    '--script',
                    '--check-cache-device',
                    '--cache-device', device]
        return cls.run_cmd(cmd)

    @classmethod
    def start_cache(cls, device, cache_id=None, cache_mode=None, cache_line_size=None, load=False, force=False):
        cmd = [cls.casadm_path,
                    '--start-cache',
                    '--cache-device', device]
        if cache_id:
            cmd += ['--cache-id', str(cache_id)]
        if cache_mode:
            cmd += ['--cache-mode', cache_mode]
        if cache_line_size:
            cmd += ['--cache-line-size', str(cache_line_size)]
        if load:
            cmd += ['--load']
        if force:
            cmd += ['--force']
        return cls.run_cmd(cmd)

    @classmethod
    def add_core(cls, device, cache_id, core_id=None, try_add=False):
        cmd = [cls.casadm_path,
                    '--script',
                    '--add-core',
                    '--core-device', device,
                    '--cache-id', str(cache_id)]
        if core_id:
            cmd += ['--core-id', str(core_id)]
        if try_add:
            cmd += ['--try-add']
        return cls.run_cmd(cmd)

    @classmethod
    def stop_cache(cls, cache_id, no_flush=False):
        cmd = [cls.casadm_path,
                    '--stop-cache',
                    '--cache-id', str(cache_id)]
        if no_flush:
            cmd += ['--no-data-flush']
        return cls.run_cmd(cmd)

    @classmethod
    def remove_core(cls, cache_id, core_id, detach=False, force=False):
        cmd = [cls.casadm_path,
                    '--script',
                    '--remove-core',
                    '--cache-id', str(cache_id),
                    '--core-id', str(core_id)]
        if detach:
            cmd += ['--detach']
        if force:
            cmd += ['--no-flush']
        return cls.run_cmd(cmd)

    @classmethod
    def flush_parameters(cls, cache_id, policy_type):
        cmd = [cls.casadm_path,
                    '--flush-parameters',
                    '--cache-id', str(cache_id),
                    '--cleaning-policy-type', policy_type]
        return cls.run_cmd(cmd)

    @classmethod
    def io_class_load_config(cls, cache_id, ioclass_file):
        cmd = [cls.casadm_path,
                    '--io-class',
                    '--load-config',
                    '--cache-id', str(cache_id),
                    '--file', ioclass_file]
        return cls.run_cmd(cmd)

# Configuration file parser

class cas_config:
    class cache_config:
        def __init__(self, line):
            values = line.split()
            self.cache_id = values[0]
            self.device = values[1]
            self.cache_mode = values[2].lower()
            self.params = dict()
            if len(values) > 3:
                for param in values[3].split(','):
                    self.params.update([param.split('=')])
            self.cores = dict()

    class core_config:
        def __init__(self, line):
            values = line.split()
            self.cache_id = values[0]
            self.core_id = values[1]
            self.device = values[2]

    def __init__(self, config_file):
        section_caches = False
        section_cores = False
        self.caches = dict()
        self.cores = list()
        try:
            with open(config_file, 'r') as conf:
                version_tag = conf.readline()
                if not re.findall(r'^version=.*$', version_tag):
                    raise ValueError('No version tag found!')
                for line in conf:
                    if not re.findall(r'^\s*#|^\s*$', line):
                        line = line.split('#')[0].rstrip()
                        if line == '[caches]':
                            section_caches = True
                            continue
                        if line == '[cores]':
                            section_caches = False
                            section_cores = True
                            continue
                        if section_caches:
                            cache = cas_config.cache_config(line)
                            self.caches[cache.cache_id] = cache
                        elif section_cores:
                            core = cas_config.core_config(line)
                            self.caches[core.cache_id].cores[core.core_id] = core
                            self.cores += [core]
        except ValueError:
            raise
        except IOError:
            raise Exception('Couldn\'t open config file')
        except:
            raise Exception('Malformed config file')

# Config helper functions

def config_start_cache(cache, load):
    casadm.start_cache(
            device = cache.device,
            cache_id = cache.cache_id,
            cache_mode = cache.cache_mode,
            cache_line_size = cache.params.get('cache_line_size'),
            load = load)

def config_configure_cache(cache):
    if cache.params.has_key('cleaning_policy'):
        casadm.flush_parameters(
                cache_id = cache.cache_id,
                policy_type = cache.params['cleaning_policy'])
    if cache.params.has_key('ioclass_file'):
        casadm.io_class_load_config(
                cache_id = cache.cache_id,
                ioclass_file = cache.params['ioclass_file'])

def config_add_core(core, attach):
    casadm.add_core(
            device = core.device,
            cache_id = core.cache_id,
            core_id = core.core_id,
            try_add = attach)

# Another helper functions

def get_caches_list():
    result = casadm.list_caches()
    return list(csv.DictReader(result.stdout.split('\n')))

def check_cache_device(device):
    result = casadm.check_cache_device(device)
    return list(csv.DictReader(result.stdout.split('\n')))[0]
